# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.plugins.factory import BaseFactory
from aiida.tools.dbimporters.baseclasses import DbImporter


def DbImporterFactory(entry_point):
    """
    Return the DbImporter plugin class for a given entry point

    :param entry_point: the entry point name of the DbImporter plugin
    """
    return BaseFactory('aiida.tools.dbimporters', entry_point)
